// Copyright 2001, Jonathan Bachrach.  See file TERMS.

import java.io.*;
import java.util.*;
import java.lang.reflect.*;
import JavaLexer;
import antlr.*;

/**
 * Java PreProcessor parser/macro expander engine
 * @author  Jonathan Bachrach
 * @author  Keith Playford
 * @version 0.9, 03OCT00
 */
public class LooseFragmentParser {
    public static JavaReservedWords reservedWords = new JavaReservedWords();
    JavaLexer lexer         = null;
    List queue              = List.nil;
    List endTokens          = List.nil;
    static boolean isTraced = false;

    void sharedInitialize (boolean trace) {
	isTraced           = trace;
    }

    void initialize (InputStream stream, boolean trace) {
	lexer = new JavaLexer(stream);
	sharedInitialize(trace);
    }

    void initialize (StringReader stream, boolean trace) {
	lexer = new JavaLexer(stream);
	sharedInitialize(trace);
    }

    LooseFragmentParser(InputStream stream, boolean trace) throws Exception {
	initialize(stream, trace);
    }

    LooseFragmentParser(String filename, boolean trace) throws Exception {
	try {
	    File file          = new File(filename);
	    InputStream stream = new FileInputStream(file);
	    initialize(stream, trace);
	} catch (Exception e) {
	    System.err.println("scanner exception: "+e);
	    e.printStackTrace();   // so we can get stack trace		
	}	    
    }

    LooseFragmentParser(StringReader stream, boolean trace) throws Exception {
	initialize(stream, trace);
    }

    public static void trace (String string) {
	if (isTraced)
	    System.out.println(string);
    }

    private void burp (int depth, String string) {
	if (isTraced) {
	    for (int i = 0; i < depth; i++)
		System.out.print("  ");
	    System.out.println(string);
	}
    }

    public static void debugOut (String name) {
	if (isTraced)
	    System.out.println(name);
    }

    public static void debugOut (String name, FragmentList fragments) {
	if (isTraced) {
	    System.out.println(name);
	    fragments.pprint(Fragment.out, 1);
	    System.out.println();
	}
    }

    private void pushEndToken (Token t) {
	endTokens = endTokens.push(t);
    }

    private Token popEndToken () {
	Token t = (Token)endTokens.getHead();
	endTokens = endTokens.getTail();
	return t;
    }

    private void enqueue (Token t) {
	trace("Enqueuing " + t);
	queue = queue.append(List.list(t));
    }

    private void enqueue (List tokens) {
	for (List ts = tokens; ts != List.nil; ts = ts.getTail())
	    enqueue((Token)ts.getHead());
    }

    private Token dequeue () {
	Token t = (Token)queue.getHead();
	trace("Dequeuing " + t);
	queue = queue.getTail();
	return t;
    }

    public Token nextToken () throws IOException {
	if (queue.isEmpty())
	    return lexer.nextToken();
	else
	    return dequeue();
    }

    public FragmentList readAtomsTo (int toType, int depth, boolean endToken) {
	burp(depth, "PARSING TO " + toType);
	try {
	    FragmentList fs = FragmentList.FragmentNil;
	    Token t         = nextToken();
	    while (t.getType() != JavaTokenTypes.EOF && t.getType() != toType) {
		Fragment f = readAtom(t, depth);
		burp(depth, "TOKEN: " + t + " FRAGMENT: " + f);
		fs = fs.fpush(f);
		t  = nextToken();
	    }
	    if (endToken)
		pushEndToken(t);
	    return (FragmentList)(fs.reverse());
	} catch (Exception e) {
	    System.err.println("scanner exception: "+e);
	    e.printStackTrace();   // so we can get stack trace		
	    return null;
	}	    
    }

    public FragmentList readNestedAtomsTo (int toType, int depth) {
	return readAtomsTo(toType, depth, true);
    }

    public FragmentList parse () {
	trace("PARSING");
	return readAtomsTo(JavaTokenTypes.EOF, 0, false);
    }

    public Fragment readNextAtom (int depth) {
	burp(depth, "ReadNextAtom");
	try {
	    Fragment atom = readAtom(nextToken(), depth);
	    trace("  read " + atom);
	    return atom;
	} catch (Exception e) {
	    System.err.println("scanner exception: "+e);
	    e.printStackTrace();   // so we can get stack trace		
	    return null;
	}	    
    }

    void lexerError (String msg) { }

    public Fragment readAtom (Token t, int depth) {
	Fragment atom = null;
	switch (t.getType()) {
	case JavaTokenTypes.EOF: 
	    atom = null; break;
	case JavaTokenTypes.SL_COMMENT: 
	    atom = readNextAtom(depth); break;
	case JavaTokenTypes.ML_COMMENT:
	    atom = readNextAtom(depth); break;
	case JavaTokenTypes.DOT:
	    atom = new DotFragment(t); break;
	case JavaTokenTypes.SEMI:
	    atom = new SemicolonFragment(t); break;
	case JavaTokenTypes.COMMA:
	    atom = new CommaFragment(t); break;
	case JavaTokenTypes.QUESTION:
	    Token nt = null;
	    try { nt = nextToken(); } catch (Exception e) { }	    
	    if (nt.getType() == JavaTokenTypes.LPAREN) 
		atom = new QuestionExpressionFragment
		    (t, nt, readNestedAtomsTo(JavaTokenTypes.RPAREN, depth + 1), popEndToken());
	    else {		
		atom = new QuestionFragment(t); 
		enqueue(nt);
	    } 
	    break;
	case JavaTokenTypes.DOTDOTDOT:
	    atom = new DotDotDotFragment(t); break;
	case JavaTokenTypes.LPAREN:
	    atom = new ParensFragment
		(t, readNestedAtomsTo(JavaTokenTypes.RPAREN, depth + 1), popEndToken()); break;
	case JavaTokenTypes.LBRACK:
	    atom = new BracketsFragment
		(t, readNestedAtomsTo(JavaTokenTypes.RBRACK, depth + 1), popEndToken()); break;
	case JavaTokenTypes.LCURLY:
	    atom = new BracesFragment
		(t, readNestedAtomsTo(JavaTokenTypes.RCURLY, depth + 1), popEndToken()); break;
	case JavaTokenTypes.HASHLCURLY:
	    Template tt = new Template
		(t, readNestedAtomsTo(JavaTokenTypes.RCURLY, depth + 1), popEndToken());
	    if (true /* raw */)
		atom = tt; 
	    else
		atom = tt.compileTemplate(0); 
	    break;
	case JavaTokenTypes.STRING_LITERAL:
	    atom = new StringFragment(t); break;
	case JavaTokenTypes.CHAR_LITERAL: 
	    atom = new CharacterFragment(t); break;
	case JavaTokenTypes.NUM_INT: 
	    atom = new IntegerFragment(t); break;
	case JavaTokenTypes.NUM_FLOAT:
	    atom = new FloatFragment(t); break;
	case JavaTokenTypes.PATVAR:
	    atom = new PatternVariableFragment(t); break;
	case JavaTokenTypes.MACROCALL:
	    System.out.println("FOUND MACROCALL");
	case JavaTokenTypes.IDENT:
	default:
	    atom = new IdentifierFragment(t); break;
	}
	return atom;
    }
}
